import puppeteer from 'puppeteer-extra';
import stealth from 'puppeteer-extra-plugin-stealth';
import axios from 'axios';
import path from 'path';
import fs from 'fs';
import cp from 'child_process';
import AdmZip from 'adm-zip';

puppeteer.use(stealth());

export interface ServerProperties {
    serverName: string;
    gamemode: "survival" | "creative";
    forceGamemode: boolean;
    difficulty: "peaceful" | "easy" | "normal" | "hard";
    allowCheats: boolean;
    maxPlayers: number;
    onlineMode: boolean;
    allowList: boolean;
    serverPort: number;
    serverPortv6: number;
    viewDistance: number;
    tickDistance: number;
    playerIdleTimeout: number;
    maxThreads: number;
    levelName: string;
    levelSeed: string;
    defaultPlayerPermissionLevel: "visitor" | "member" | "operator";
    texturepackRequired: boolean;
    contentLogFileEnabled: boolean;
    compressionThreshold: number;
    serverAuthoritativeMovement: "client-auth" | "server-auth" | "server-auth-with-rewind";
    playerMovementScoreThreshold: number;
    playerMovementActionDirectionThreshold: number;
    playerMovementDistanceThreshold: number;
    playerMovementDurationThresholdInMs: number;
    correctPlayerMovement: boolean;
    serverAuthoritativeBlockBreaking: boolean;
    emitServerTelemetry: boolean;
}

class BedrockServer {
    proc: cp.ChildProcess;

    public static server_properties_default: ServerProperties = {
        serverName: "Bedrock Server",
        allowCheats: false,
        allowList: false,
        compressionThreshold: 1,
        difficulty: 'normal',
        contentLogFileEnabled: false,
        correctPlayerMovement: false,
        defaultPlayerPermissionLevel: 'member',
        forceGamemode: false,
        gamemode: 'survival',
        levelName: 'A bedrock world',
        levelSeed: '',
        maxPlayers: 20,
        maxThreads: 8,
        onlineMode: true,
        playerIdleTimeout: 30,
        playerMovementActionDirectionThreshold: 0.85,
        playerMovementDistanceThreshold: 0.3,
        playerMovementDurationThresholdInMs: 500,
        playerMovementScoreThreshold: 20,
        serverAuthoritativeMovement: 'server-auth',
        serverAuthoritativeBlockBreaking: false,
        serverPort: 19132,
        serverPortv6: 19133,
        texturepackRequired: false,
        tickDistance: 4,
        viewDistance: 32,
        emitServerTelemetry: false
    };

    public static server_properties_data: string = `server-name=%%{server-name}
gamemode=%%{gamemode}
force-gamemode=%%{force-gamemode}
difficulty=%%{difficulty}
allow-cheats=%%{allow-cheats}
max-players=%%{max-players}
online-mode=%%{online-mode}
allow-list=%%{allow-list}
server-port=%%{server-port}
server-portv6=%%{server-portv6}
view-distance=%%{view-distance}
tick-distance=%%{tick-distance}
player-idle-timeout=%%{player-idle-timeout}
max-threads=%%{max-threads}
level-name=%%{level-name}
level-seed=%%{level-seed}
default-player-permission-level=%%{default-player-permission-level}
texturepack-required=%%{texturepack-required}
content-log-file-enabled=%%{content-log-file-enabled}
compression-threshold=%%{compression-threshold}
server-authoritative-movement=%%{server-authoritative-movement}
player-movement-score-threshold=%%{player-movement-score-threshold}
player-movement-action-direction-threshold=%%{player-movement-action-direction-threshold}
player-movement-distance-threshold=%%{player-movement-distance-threshold}
player-movement-duration-threshold-in-ms=%%{player-movement-duration-threshold-in-ms}
correct-player-movement=%%{correct-player-movement}
server-authoritative-block-breaking=%%{server-authoritative-block-breaking}
emit-server-telemetry=%%{emit-server-telemetry}`;

    public static async getLatestDownloadURL(url: string = 'https://www.minecraft.net/en-us/download/server/bedrock'): Promise<string> {
        const browser = await puppeteer.launch({
            headless: false,
            args: ['--no-sandbox', '--disable-setuid-sandbox'],
            defaultViewport: {
                width: 1920,
                height: 1080
            }
        });

        const [page] = await browser.pages();

        page.setDefaultNavigationTimeout(0);

        const session = await page.target().createCDPSession();
        const { windowId } = await session.send('Browser.getWindowForTarget');
        await session.send('Browser.setWindowBounds', { windowId, bounds: { windowState: 'minimized' } });

        await page.goto(url);

        const el = await page.waitForSelector('#main-content > div > div > div.page-section-container.aem-GridColumn.aem-GridColumn--default--12 > div > div > div > div.server-card.aem-GridColumn.aem-GridColumn--default--12 > div > div > div > div:nth-child(1) > div.card-footer > div > a');

        const value = (await (await el.getProperty('href')).jsonValue() as string).split('/').pop();

        await page.close();
        await browser.close();

        const os = process.platform === 'win32' ? 'win' : process.platform;
        if (!['win', 'linux'].includes(os))
            throw new Error(`Unsupported OS: ${os}`);

        const dlurl = `https://minecraft.azureedge.net/bin-${os}/${value}`;

        return dlurl;
    }

    public static camelcaseToDashCase(str: string): string {
        return str.replace(/([a-z])([A-Z])/g, '$1-$2').toLowerCase();
    }

    public static createServerProperitesData(serverProperties: ServerProperties = BedrockServer.server_properties_default): string {
        let data = BedrockServer.server_properties_data;

        for (const [key, value] of Object.entries(serverProperties)) {
            const regex = new RegExp(`%%\\{${BedrockServer.camelcaseToDashCase(key)}\\}`, 'g');

            data = data.replace(regex, value);
        }

        return data;
    }

    constructor(readonly folder: string) { }

    public async initServerFolder(zipurl: string, properties?: ServerProperties, files?: {
        [key: string]: string | Buffer;
    }): Promise<void>;
    public initServerFolder(zipdata: Buffer, properties?: ServerProperties, files?: {
        [key: string]: string | Buffer;
    }): void;
    public initServerFolder(zipdata: string | Buffer, properties: ServerProperties = BedrockServer.server_properties_default, files: {
        [key: string]: string | Buffer;
    } = {}): Promise<void> | void {
        if (fs.existsSync(this.folder))
            try {
                fs.rmSync(this.folder, {
                    recursive: true,
                    force: true
                });
            } catch (e: any) {
                if (e.code == 'ENOPERM') {
                    console.log('Permission denied while trying to remove old folder.');

                    process.exit(1);
                }
            }

        fs.mkdirSync(this.folder, { recursive: true });

        if (!Buffer.isBuffer(zipdata))
            return new Promise<void>(async resolve => {
                zipdata = await (await axios.get(zipdata as string, {
                    responseType: 'arraybuffer'
                })).data;

                const zip = new AdmZip(zipdata);

                const zipEntries = zip.getEntries();

                for (const entry of zipEntries) {
                    console.log(`\x1b[32m[${entry.name}] Extracting...\x1b[0m`);
                    if (entry.isDirectory) {
                        continue;
                    }

                    const entryName = entry.entryName;
                    const entryData = zip.readFile(entryName);
                    const entryPath = path.join(this.folder, entryName);
                    const entryDir = path.dirname(entryPath);
                    if (!fs.existsSync(entryDir)) {
                        fs.mkdirSync(entryDir, { recursive: true });
                    }

                    fs.writeFileSync(entryPath, entryData);

                    console.log(`\x1b[32m[${entry.name}] Extracted to ${entryPath}\x1b[0m`);
                }

                for (const [filename, data] of Object.entries(files)) {
                    fs.writeFileSync(filename, data);
                }

                this.changeServerProperties(properties);

                resolve();
            });

        const zip = new AdmZip(zipdata);

        const zipEntries = zip.getEntries();

        for (const entry of zipEntries) {
            console.log(`\x1b[32m[${entry.name}] Extracting...\x1b[0m`);
            if (entry.isDirectory) {
                continue;
            }

            const entryName = entry.entryName;
            const entryData = zip.readFile(entryName);
            const entryPath = path.join(this.folder, entryName);
            const entryDir = path.dirname(entryPath);
            if (!fs.existsSync(entryDir)) {
                fs.mkdirSync(entryDir, { recursive: true });
            }

            fs.writeFileSync(entryPath, entryData);

            console.log(`\x1b[32m[${entry.name}] Extracted to ${entryPath}\x1b[0m`);
        }

        this.changeServerProperties(properties);

        for (const [filename, data] of Object.entries(files)) {
            fs.writeFileSync(filename, data);
        }
    }

    public changeServerProperties(properties: ServerProperties): void {
        const data = BedrockServer.createServerProperitesData(properties);

        fs.writeFileSync(path.join(this.folder, "server.properties"), data);
    }

    public start(): cp.ChildProcess {
        const proc = cp.spawn(path.resolve(path.join(this.folder, `bedrock_server${process.platform === 'win32' ? ".exe" : ""}`)), [], {
            cwd: this.folder,
            stdio: 'pipe'
        });

        this.proc = proc;

        return proc;
    }

    public stop(code: NodeJS.Signals = 'SIGINT'): number {
        this.proc.kill(code);

        return this.proc.exitCode;
    }
}

export default BedrockServer;